from ..search import things

from ..util.fonts import platform

Agent = things.Agent
UtilityBasedAgent = things.UtilityBasedAgent

class Hospital(things.Thing):
    def __repr__(self):
        return "🏥" if platform != "darwin" else "+"

class House(things.Thing):
    def __repr__(self):
        return "🏠" if platform != "darwin" else "^"

class Optimiser(things.Agent):
    def __repr__(self):
        return "📈"


from ..agent.environment import XYEnvironment
from ..search.things import *
from numpy import inf, min, max

from copy import deepcopy

class Tile(Thing):
    def __init__(self, player=None):
        self.player = player
        self.min_val, self.max_val = inf, -inf
    
    def __repr__(self):
        return self.player if self.player else " "


class TicTacToeGame(XYEnvironment):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, width=3, height=3, **kwargs)
        self.board = [[Tile() for j in range(3)] for i in range(3)]
        self.in_play = True

    @property
    def is_done(self):
        if len(self.agents) == 0: return True
        return not self.in_play

    def step(self):
        if self.is_done: return
        def get_position():
            print(self)
            i = int(input("choose move row [1-3]"))-1
            j = int(input("choose move column [1-3]"))-1
            return i,j
        i,j = get_position()
        placed = False
        if not self.board[i][j].player:
            self.board[i][j].player = "X" if self.agent.player == "O" else "O"
            placed = True
        while(not placed):
            print("not a valid move !")
            i,j = get_position()
            if (i < 0 or i > 2 or j < 0 or j > 2):
                continue
            if not self.board[i][j].player:
                self.board[i][j].player = "X" if self.agent.player == "O" else "O"
                placed = True
        print(self)
        super().step()
    
    def __repr__(self):
        board = ""
        for i in (0,1):
            board += f"_{self.board[i][0]}_|_{self.board[i][1]}_|_{self.board[i][2]}_"
            board += "\n"
        board += f" {self.board[2][0]} | {self.board[2][1]} | {self.board[2][2]} "
        return board[:-1]
    
    def add_agent(self, agent, player="X"):
        if not isinstance(agent, Agent):
            raise TypeError(f"{self}: {agent} is not an Agent")
        self.agents.add(agent)
        self.agent = agent
        self.agent.player = "X"

    def percept(self, agent):
        return self.board

    def execute_action(self, agent, action):
        command, state = action
        print(action)
        match command:
            case "move":
                self.board = agent.move(state)
            case "done":
                self.in_play = False


class MinimaxAgent(UtilityBasedAgent):

    def to_move(self, state):
        NotImplemented

    def moves(self, state):
        NotImplemented

    def score(self, state):
        NotImplemented
    
    def minimax_utility(self, state):
        match self.to_move(state):
            case "min":
                return min(
                    [self.minimax_utility(move) for move in self.moves(state)])
            case "max":
                return max(
                    [self.minimax_utility(move) for move in self.moves(state)])
            case "terminal":
                return self.score(state)

    def utility(self, action):
        _, state = action
        return self.minimax_utility(state)
    
    def program(self,percepts):
        print(f"{self}: thinking ...")
        state = percepts
        if self.to_move(state) == "terminal":
            return ("done", state)
        
        max_objective = -inf
        action = self.maximise_utility(
            [("move", move) for move in self.moves(state)])
        
        return action
    

class TicTacToeAgent(MinimaxAgent):
    def __repr__(self):
        if hasattr(self, "player"):
            return self.player
        return super().__repr__()
        
    def score(self, state):
        draw = True  # check
        index = lambda x,i,j: x[i][j]
        def is_win(check):
            if all(index(state, *idx).player == tile.player for idx in check):
                return True
            return False
        
        for i in range(3):
            for j in range(3):
                tile = index(state,i,j)
                if not tile.player:  # is empty
                    draw = False
                    continue
                check = [
                    [((i_+1)%3, j) for i_ in range(i, i+2)],
                    [(i,(j_+1)%3) for j_ in range(j, j+2)]]
                if i==j:
                    check.append([((i_+1)%3, (i_+1)%3) for i_ in range(i, i+2)])
                if i+j == 2:
                    check.append([((i_+1)%3, (2-(i_+1)%3)%3) for i_ in range(i, i+2)])
                if any(is_win(idxs) for idxs in check):
                    return 1 if tile.player == self.player else -1
        
        return 0 if draw else None

    def move(self, state):
        return state
    
    def moves(self, state):
        to_move = self.to_move(state)
        if self.player == "X":
            player = "X" if to_move == "max" else "O"
        else:
            player = "O" if to_move == "max" else "X"
        possible_moves = []
        for i in range(3):
            for j in range(3):
                if not state[i][j].player:
                    move = deepcopy(state)#.copy()
                    move[i][j] = Tile(player)
                    possible_moves.append(move)
        return possible_moves

    def to_move(self, state):
        if self.score(state) is not None:
            return "terminal"
        moves_made = {"X":0, "O":0}
        for i in range(3):
            for j in range(3):
                if state[i][j].player:
                    moves_made[state[i][j].player] += 1
        match self.player:
            case "X":
                return "max" if moves_made["X"] < moves_made["O"] else "min"
            case "O":
                return "max" if moves_made["O"] < moves_made["X"] else "min"


if __name__ == "__main__":
    env = TicTacToeGame()
    env.add_agent(TicTacToeAgent())
    env.run()


"""ENGINE.PY
This contains the code for the base engine
"""
import warnings
import time
from datetime import datetime
from collections.abc import Iterable, Callable
from typing import override

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pygame

if __name__ == "__main__":  # being run as script
    from co2114.util import colours, fonts
    COLOR_BLACK, COLOR_WHITE = colours.COLOR_BLACK, colours.COLOR_WHITE
else:
    from .util.colours import COLOR_BLACK, COLOR_WHITE
    from .util import fonts
    
TEXT_FONT = fonts._get_text_font_unsafe()  # default text font

## GLOBALS
DEFAULT_FPS:int = 30  # Render frames per second
DEFAULT_LPS:int = 2   # Environment steps per second

class BaseEngine:
    """ Base Engine Class 
            
            Provides basic PyGame setup and event loop
    """
    size = width, height = 600, 400  # default window size

    def __init__(self) -> None:
        """ Initialise Base Engine """
        print(f"Creating instance of {self.name} ({self.__class__.__name__})")

        pygame.init() # initialise pygame

        self.screen = pygame.display.set_mode(self.size)
        pygame.display.set_caption(f"{self.name} (co2114)")
        self._font = pygame.font.SysFont(TEXT_FONT, 28)  # default system font
        self._running:bool = True


    @property
    def name(self) -> str:
        """ Get engine name """
        return self.__class__.__name__
    
    @property
    def isrunning(self) -> bool:
        """ Check if engine is running """
        return self._running
    

    def on_event(self, event:pygame.event.Event) -> None:
        """ Event handler """
        if event.type == pygame.QUIT:
            self._running = False


    def cleanup(self) -> None:
        """ Cleanup engine on exit """
        print(f"Quiting {self.name}")
        pygame.quit()


    def _run(self, *render_sequence:Callable) -> None:
        """ Main event loop 
        
        :param *render_sequence: sequence of functions to call each loop
        """
        while self.isrunning:
            for event in pygame.event.get():
                self.on_event(event)  # handle events
            for fcn in render_sequence:  # call each function
                fcn()
        self.cleanup()  # cleanup on exit


class Engine(BaseEngine):
    """Engine Class
        Provides basic framerate and looprate control"""
    
    def __init__(self,
                 fps:int = DEFAULT_FPS,
                 lps:int = DEFAULT_LPS,
                 dims:tuple[int,int]|None = None) -> None:
        """ Initialise Engine
        
        :param fps: frames per second (render rate)
        :param lps: loops per second (process rate)
        :param dims: optional dimensions tuple (width, height)
        """
        # set dimensions if provided
        if dims and isinstance(dims, Iterable) and len(dims)==2:
            self.size = self.width, self.height = dims
    
        super().__init__()  # initialise base engine

        self._framerate:int = fps if isinstance(fps, int) else DEFAULT_FPS
        self._looprate:int = lps if isinstance(lps, int) else DEFAULT_LPS
        self._t0, self._l0 = time.time(), time.time()  # timing count


    def _update(self) -> None:
        """ Processing loop internals """
        # frame limiter
        if self._framerate is None or not isinstance(self._framerate, int):
            self._framerate = DEFAULT_FPS  # corrective measures
        mspf = (1000/self._framerate)  # ms per frame
        if (time.time() - self._t0) < mspf:  # if faster than framerate
            time.sleep((mspf - (time.time()-self._t0))/1000)  # sleep
        self._t0 = time.time()  # update timer

        # main render loop
        if self._looprate is None or not isinstance(self._looprate, int):
            self._looprate = DEFAULT_LPS  # corrective measures
        spf = 1/self._looprate  # s per frame
        if(time.time() - self._l0) > spf:  # if enough time passed
            self._l0 = time.time()  # update timer
            self.update()  # run main process update


    def _render(self) -> None:
        """ Render loop internals """
        self.screen.fill(COLOR_BLACK)  # write black to buffer
        self.render()  # run main render loop
        pygame.display.flip()  # flip buffer


    def update(self) -> None:
        """ Main process loop, needs override """
        raise NotImplementedError

    def render(self) -> None:
        """ Main render loop, needs override """
        raise NotImplementedError
    
    
    def run(self) -> None:
        """ Run the engine """
        super()._run(self._update, self._render)


class App(Engine):
    """ Base App Class

            Provides default name and run method for GUI apps
    """
    def __init__(self, *args, name:str | None = None, **kwargs) -> None:
        """ Initialise App
        
        :param    *args: positional arguments for Engine
        :param     name: optional name for the app
        :param **kwargs: keyword arguments for Engine
        """
        if name is not None: self._name = name  # set name if provided
        super().__init__(*args, **kwargs)  # initialise engine

    
    @classmethod
    def run_default(cls) -> None:
        """ Run the app with default parameters """
        app = cls()
        app.run()


    @property
    def name(self) -> str:
        """ Get app name """
        return self._name if hasattr(self, "_name") else super().name


class EmptyApp(App):
    """ Dummy Class for testing """
    pass

#######################
class ClockApp(App):
    """ Example PyGame App """
    
    # size = width, height = 600, 400  # uncomment to override

    def __init__(self) -> None:
        """ Initialise Clock App """
        super().__init__(fps=60, lps=60)  # initialise app
        self._font = pygame.font.SysFont(TEXT_FONT, 128)  # override font
        self.t = datetime.now()  # current time

    # @override
    def update(self) -> None:
        """ Main process loop

                Gets current system time
        """
        self.t = datetime.now()  # get current time
    
    # @override
    def render(self) -> None:
        """ Main render loop 

                Writes time to screen
        """
        # render time string
        renderTime = self._font.render(
            self.t.strftime("%H:%M:%S.%f")[:-5],
            True,
            COLOR_WHITE)
        rect = renderTime.get_rect()
        # blit to screen centered
        self.screen.blit(
            renderTime, 
            (self.width//2 - rect.width//2, 
             self.height//2 - rect.height//2))


####################################
if __name__ == "__main__":
    """ Run ClockApp if engine.py is run as script """
    print("Running engine.py as script.")
    ClockApp.run_default()


from sys import platform


def _get_text_font_unsafe():
    """ Returns None which should cause Pygame to fall to default """
    return None

def _get_symbolic_font_unsafe():
    """ Returns expected system font """
    match platform:
        case "linux":
            return "Noto Color Emojis" # TODO test
        case "darwin":
            return None #Apple Color Emoji" # TODO test
        case "win32":
            return "segoeuiemoji" 
        case _:
            return None



from __future__ import annotations

import math
import warnings
import json

from typing import Iterable, Iterator, override, Callable, Collection, TypeVar

from matplotlib import pyplot as plt
from matplotlib.figure import Figure  # for type hinting

from ..agent.environment import Environment
from .things import Thing, Agent, UtilityBasedAgent, Action

Numeric = int |  float | None
Location = Thing | Iterable
Label = str
# Weight = Numeric

# Example graph dictionary representing UK cities and distances between them
UK_CITIES_GRAPH:dict[str, Iterable[str | tuple[str,str] | int]] = {
    "nodes": ["Edinburgh", "Glasgow", "Manchester", "Liverpool",
             "Newcastle", "York", "Sheffield", "Leicester", 
             "London", "Bath", "Bristol", "Exeter", "Cardiff", "Birmingham"],
    "edges": [("Edinburgh", "Glasgow"), ("Glasgow", "Newcastle"),
             ("Edinburgh", "Newcastle"), ("Newcastle", "York"),
             ("Glasgow", "Manchester"), ("Manchester", "Liverpool"),
             ("York", "Sheffield"), ("York", "Leicester"),
             ("Sheffield", "Leicester"), ("Sheffield", "Birmingham"),
             ("Sheffield", "London"), ("Manchester", "Sheffield"),
             ("Manchester", "Birmingham"),  ("Birmingham", "Liverpool"),
             ("Birmingham", "Cardiff"), ("Leicester", "London"),
              ("Birmingham", "Leicester"),
             ("Birmingham", "London"), ("Birmingham", "Bristol"),
             ("London", "Bristol"), ("Cardiff", "Bristol")],
    "weights": [2,4,1,1,4,1,1,2,1,3,3,2,1,2,5,1,2,2,3,2,1]          
}


class Node(Thing):
    """ A node in a graph structure."""
    def __init__(self, label:Label = "") -> None:
        """ Create a node with given label. """
        self.label:Label = label  # identifier for node
        self.neighbours:set[Node] = set()  # set of neighbouring nodes
        self.weights:dict[Node, Numeric] = {}  # mapping of neighbouring nodes to edge weights
        self.location:Location | None = None  # optional location attribute

    @override
    def __repr__(self) -> str:
        """ String representation of node, returning its label. """
        return self.label
    

    def add_neighbour(self, node:Node, weight:Numeric = None) -> None:
        """ Add a neighbouring node with optional edge weight. 
        
        :param   node: Node to be added as a neighbour.
        :param weight: Weight of edge to neighbour (default None).
        """
        if not isinstance(node, Node):  # type check
            raise TypeError(f"{self}: {node} is not a Node")
        
        if node not in self.neighbours:  # avoid duplicate edges
            self.neighbours.add(node)
            self.weights[node] = weight
            node.add_neighbour(self, weight)  # undirected graph



Edge = tuple[Node, Node, Numeric]  # edge type alias


class Graph:
    """ A graph structure comprising nodes and edges. """
    def __init__(self) -> None:
        """ Create an empty graph. """
        self.nodes:set[Node] = set()

    
    def __iter__(self) -> Iterator[Node]:
        """ Iterate over nodes in the graph. """
        return iter(self.nodes)


    def add_node(self, node:Node) -> None:
        """ Add a node and its neighbours to the graph.
        
        :param node: Node to be added to the graph.
        """
        if not isinstance(node, Node):  # type check
            raise TypeError(f"{self}: {node} is not a Node")
        
        if node not in self.nodes:  # avoid duplicate nodes
            self.nodes.add(node)
            for neighbour in node.neighbours:  # add neighbours recursively
                self.add_node(neighbour)


    def plot_nodes(
            self,
            condition:Callable | None = None,
            init:Node | Label | None = None,
            labels:Iterable[Label] | None = None) -> Figure:
        """ Plot the nodes in the graph
        
        :param condition: Function to filter nodes to be plotted. If None, all nodes are plotted.
        :param init:     Node to use as the root of the plot. If None, a random node is used.
        :param labels:   Iterable of labels for the nodes. If None, node labels are used.
        """

        # filter nodes based on condition
        nodes = self.nodes if condition is None else \
            {node for node in self.nodes if condition(node)}
        
        # handle empty graph
        if len(nodes) == 0:
            return plt.figure(figsize=(8,8))
        
        # determine initial node for layout
        _init = Node()  # placeholder
        if not isinstance(init, Node):
            if init is None:
                _init = next(iter(nodes))  # random node
            elif isinstance(init, Label): # label provided
                _flag = True
                for node in nodes: # find node with matching label
                    if init == node.label:
                        _init = node
                        _flag = False
                        break
                if _flag:
                    _init = next(iter(nodes))
            else:
                raise TypeError(f"{self}: {init} is not a Node or label")
        elif init not in nodes: # node not in filtered set
            _init = next(iter(nodes)) # random node
        
        locs = {_init: (0.,0.)} # locations of nodes
        tree = {0: {_init}} # nodes at each distance from init
        dist = 1  # dist from init
        tree[dist] = _init.neighbours & nodes  # set intersection
        for i, node in enumerate(tree[dist]):
            locs[node] = (dist, i-0.5*len(tree[dist])) # initial layout

        while len(tree[dist]) > 0:  # expand tree
            tree[dist+1] = set()
            # find new nodes at dist+1 from init
            for node in tree[dist]:
                neighbours = node.neighbours & nodes # set intersection
                for _node in neighbours:
                    if _node not in locs and _node not in tree[dist+1]:
                        tree[dist+1].add(_node)
            dist += 1  # increment distance
            # layout new nodes
            for i, node in enumerate(tree[dist]):
                offset = i-0.5*len(tree[dist])
                locs[node] = (dist + 0.1*offset, offset)
        
        edges:set[Edge] = set()
        for node in nodes:
            for _node in (node.neighbours & nodes):
                forward:Edge = (node, _node, node.weights[_node])
                backward:Edge = (_node, node, _node.weights[node])
                if forward not in edges and backward not in edges:
                    edges.add(forward)

        fig, ax = plt.subplots(figsize=(8,8))
        for node, loc in locs.items():  # plot nodes and labels
            name = str(node) if labels is None else labels[node]
            ax.plot(*loc, 'ko', ms=20)
            ax.text(loc[0]+0.1, loc[1]+0.1, name)
        
        for edge in edges:  # plot edges and weights
            a,b, weight = locs[edge[0]], locs[edge[1]], edge[2]
            ax.plot(*[[i,j] for i,j in zip(a,b)],'k')
            if weight is not None:
                ax.text(*[i+(j-i)/2 for i,j in zip(a,b)], str(weight))
        ax.axis("off")  # turn off axes
        return fig



class GraphEnvironment(Environment):
    """ An environment based on a graph structure. """
    def __init__(self, graph:Graph | None = None, *args, **kwargs) -> None:
        """ Create a graph environment."""
        super().__init__(*args, **kwargs)  # initialize base Environment
        self.graph = graph if isinstance(graph, Graph) else Graph() 


    def add_node(self, node:Node) -> None:
        """ Add a node to the graph environment.
        
        :param node: Node to be added to the graph.
        """
        self.graph.add_node(node)


    @override
    def percept(self, agent:Agent) -> list[tuple[Node, Numeric]]:
        """ Return the percepts for the agent based on its location in the graph."""
        node:Node = agent.location  # type: ignore
        if len(node.weights) == 0:
            return [(n, 1) for n in node.neighbours]
        else:
            return [(n, node.weights[n]) for n in node.neighbours]


    @override
    def add_agent(self, 
                  agent:Agent, 
                  location:Location | None = None,
                  node:Node | None = None) -> None:
            """ Add an agent to the graph environment at a specified location or node.
            
            :param agent:    Agent to be added to the environment.
            :param location: Location where the agent should be added (default None).
            :param node:     Node where the agent should be added (default None).
            """
            if not isinstance(agent, Agent):  # type check
                raise TypeError("f{self}: {agent} is not an Agent")
            
            if node is not None:  # assign to specified node
                if not isinstance(node, Node):
                    print(f"{self}: {node} is not a valid Node")

                if node in self.graph:  
                    super().add_agent(agent, node)
                else:
                    print(f"{self}: {node} is not in graph")

            elif location is not None:  # assign to node with matching location
                _flag = True

                for node in self.graph:  # find node with matching location
                    if location == node.location:
                        super().add_agent(agent, node)
                        _flag = False
                        break
                    
                if _flag:  # location not found
                    print(f"{self}: {location} was not found in environment")

            else:  # default to random node
                super().add_agent(agent)


    def show(self, *args, **kwargs):
        """ Show a plot of the graph environment. """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.graph.plot_nodes(*args, **kwargs).show()


    @classmethod
    def from_dict(cls,
                  graph_dict: dict[str, Iterable[str | tuple[str,str] | int]]
                  )-> GraphEnvironment:
        """ Create a GraphEnvironment from a dictionary representation of a graph."""
        if "vertices" not in graph_dict and "edges" not in graph_dict:
            raise ValueError(f"No vertices in json string {graph_dict}")
        
        if "edges" not in graph_dict: 
            raise ValueError(f"No edges in json string {graph_dict}")
        
        vertices = graph_dict[  # support 'nodes' or 'vertices'
                            'vertices' if 'vertices' in graph_dict else 'nodes']
        edges = graph_dict['edges']

        for edge in edges:
            if len(edge) != 2:  # type: ignore
                raise ValueError(
                        f"Edges must comprise two nodes, {edge} does not")
            if any(v not in vertices for v in edge):  # type: ignore
                raise ValueError(
                    f"Edges must map between valid vertices, {edge} does not")
        
        nodes = {v: Node(f"{v}") for v in vertices}
        
        if 'weights' in graph_dict:
            weights = graph_dict['weights']
        else:
            weights = [None]*len(edges) # type: ignore
        for edge, weight in zip(edges, weights):
            a, b = edge  # type: ignore
            nodes[a].add_neighbour(nodes[b], weight)  # add edge with weight


        # create environment and add nodes
        environment = cls()
        for _,node in nodes.items():
            environment.add_node(node)

        return environment
    
    @classmethod
    def from_json(cls, json_str: str) -> GraphEnvironment:
        """ Create a GraphEnvironment from a JSON string representation 
        of a graph."""
        return cls.from_dict(json.loads(json_str))


class ShortestPathEnvironment(GraphEnvironment):
    """ A graph environment for shortest path finding agents. """
    def get_node(self, node: Node | Label) -> Node:
        """ Retrieve a node from the graph by Node or label.

        :param node: Node or label of the node to retrieve.
        """
        if isinstance(node, Node):
            assert node in self.graph
        else:
            for vertex in self.graph:
                if node == vertex.label:
                    node = vertex
                    break
            assert isinstance(node, Node)
        return node
    

    @override
    def add_agent(self,
                  agent:ShortestPathAgent,
                  init: Node | Label,
                  target: Node | Label) -> None:
        """ Add a shortest path agent to the environment.
        
        :param agent:  Agent to be added to the environment.
        :param init:   Initial node or label for the agent.
        :param target: Target node or label for the agent, if applicable.
        """
        # retrieve nodes from labels if necessary
        init = self.get_node(init)
        target = self.get_node(target)
        # add agent to environment
        super().add_agent(agent, node=init)
        # initialise agent with init and target nodes
        agent.initialise(init, target)


    @property  # override
    def is_done(self) -> bool:
        """ Check if all agents have delivered their paths. """
        return (hasattr(self, "delivered") and self.delivered)


    @override
    def execute_action(self,
                       agent: ShortestPathAgent,
                       action: tuple[str, Node]) -> None:
        """ Execute an action for the shortest path agent.
        
        Commands:
        - "explore": move to a neighbouring node
        - "deliver": deliver the shortest path to the target node

        :param  agent:  Agent executing the action.
        :param action: Action to be executed, format (command, node).
        """
        command, node = action
        match command:
            case "explore":
                agent.explore(node)
            case "deliver":
                if not hasattr(self, "shortest_path"):
                    self.shortest_path = {}
                self.shortest_path[agent.init] = {node: agent.deliver(node)}
                self.delivered = True


class ShortestPathAgent(UtilityBasedAgent):
    """ An agent that finds the shortest path in a graph environment.
    
    Needs implementation of methods:
        `program(self, percepts:list[tuple[Node, Numeric]]) -> tuple[str, Node]`
        `utility(action:tuple[str, Node]) -> Numeric`
        `explore(node:Node) -> None`
        `deliver(node:Node) -> tuple[list[Node], Numeric]`
    
    Has attributes:
        `init`:   Initial node of the agent.
        `target`: Target node of the agent.
        `location`: Current location of the agent.
        `dist`:   Dictionary mapping nodes to their distance from the initial node.
        `prev`:   Dictionary mapping nodes to their previous node in the shortest path.
    """
    def __init__(self):
        """ Initialise the shortest path agent. """
        super().__init__()

        # attributes for shortest path finding
        self.__init = self.__target = None
        self.location:Node | None = None
        self.dist:dict[Node, Numeric] = {}
        self.prev:dict[Node, Node | None] = {}
        self.visited:set[Node] = set()


    @property
    def at_goal(self) -> bool:
        """ Check if the agent is at the target node. """
        return self.location is self.target
    

    @property
    def init(self) -> Node | None:
        """ Get the initial node of the agent. """
        return self.__init
    

    @property
    def target(self) -> Node | None:
        """ Get the target node of the agent. """
        return self.__target


    def initialise(self, node:Node, target:Node | None) -> None:
        """ Initialise the agent with its starting and target nodes.
        
        :param node:   Initial node for the agent.
        :param target: Target node for the agent, if applicable.
        """
        self.__init = self.location = node
        self.__target = target
        self.dist[node] = 0  # distance from initial node
        self.prev[node] = None  # previous node in path


    def explore(self, node:Node) -> None:
        """ Explore a new node. The actuator of the agent. Needs to be implemented.
        
        :param node: Node to be explored.
        """
        raise NotImplementedError
    

    def deliver(self, node:Node) -> tuple[list[Node], Numeric]:
        """ Deliver the shortest path to the specified node, 
        and distance to node. Needs to be implemented
        
        :param            node: Node to deliver the shortest path to.
        :return path, distance: Tuple of (shortest path as list of nodes, distance to node).
        """
        raise NotImplementedError


import numpy as np
import random
from ...agent.things import Thing, Agent
from collections.abc import Callable, Iterable
from copy import deepcopy

from .util import alldiff

__all__ = [
    'util',
    'CSPAgent',
    'ConstraintSatisfactionProblem',
    'Variable',
    'Factor']

class CSPAgent(Agent):
    def __repr__(self):
        return "🤖"
    
    def program(self, percept):
        pass

    def solve(self):
        raise NotImplementedError


class ConstraintSatisfactionProblem:
    def __init__(self, variables, constraints):
        self.__variables = variables
        self.__constraints = constraints

    @property
    def variables(self):
        return self.__variables
    
    @property
    def domains(self):
        return {variable: variable.domain for variable in self.variables}

    @property
    def constraints(self):
        return self.__constraints
    
    @property
    def is_consistent(self):
        return all(constraint.is_satisfied for constraint in self.constraints)
    
    @property
    def is_complete(self):
        return all(variable.is_assigned for variable in self.variables) \
                    and self.is_consistent
    
    @property
    def arcs(self):
        return [arc for constraint in self.constraints if constraint.is_binary
                        for arc in constraint.arcs]
    
    def __repr__(self):
        return str({v.name: v for v in self.variables})
    

class __Variable(Thing):
    @property
    def value(self):
        return self.__value
    
    @property
    def is_assigned(self):
        return hasattr(self, '__value') and self.__value is not None

    def __eq__(self, x):
        return (x == self.value)
    
    def __ne__(self, x):
        return (x != self.value)
    
    def __lt__(self, x):
        return self.value < x if self.is_assigned else True

    def __gt__(self, x):
        return self.value > x if self.is_assigned else True
    
    def __le__(self, x):
        return self.__eq__(x) or self.__lt__(x)
    
    def __ge__(self, x):
        return self.__eq__(x) or self.__gt__(x)
    
    def __add__(self, x):
        return self.value + x if self.is_assigned else self
    
    def __sub__(self, x):
        return self.value - x if self.is_assigned else self

    def __mul__(self, x):
        return self.value * x if self.is_assigned else self
    
    def __truediv__(self, x):
        return self.value / x if self.is_assigned else self
    
    def __rtruediv__(self, x):
        return x / self.value if self.is_assigned else self
    
    def __mod__(self, x):
        return self.value % x if self.is_assigned else self

    def __or__(self, x):
        return self.value | x if self.is_assigned else self
    
    def __and__(self, x):
        return self.value & x if self.is_assigned else self
    
    def __pow__(self, x):
        return self.value ** x if self.is_assigned else self

    def __neg__(self):
        return -self.value if self.is_assigned else self

    def __truediv__(self, x):
        return self.value/x if self.is_assigned else self
    
    def __floordiv__(self, x):
        return self.value//x if self.is_assigned else self
    
    def __abs__(self):
        return abs(self.value) if self.is_assigned else self


class Variable(__Variable):
    def __init__(self, domain, name=None):
        self.__domain = deepcopy(domain)  # domain belongs only to one variable
        self.__value = None
        self.name = name
        self.__hash = random.random()

    def __hash__(self):
        return hash(self.__hash)

    @property
    def is_assigned(self):
        return self.__value is not None

    @property
    def value(self):
        return self.__value
    
    @value.setter
    def value(self, x):
        if x in self.domain or x is None:
            self.__value = x
        else:
            raise ValueError(f"{self.name}: {x} not in domain {self.domain}")

    @property
    def domain(self):
        return self.__domain
    
    def __repr__(self):
        return f"{self.name}({str(self.value) if self.is_assigned else '?'})"


class Factor(Thing):
    def __init__(self, constraint, variables):
        assert isinstance(constraint, Callable), "constraint must be callable"
        self.__function = constraint
        if not isinstance(variables, Iterable):
            variables = [variables]
        self.__variables = np.array(variables).ravel().tolist()
        if self.__xnary < 1:
            raise ValueError(f"{self}: number of variables must be >1")

    def __call__(self, *args, **kwargs):
        return self.__function(*args, **kwargs)

    def __iter__(self):
        return iter(self.__variables)
    
    def __contains__(self, variable):
        """
            overrides in keyword behaviour because variable equality 
            is checked by value rather than hash
        """
        return any(variable is _variable for _variable in self)
        

    @property
    def is_satisfied(self):
        if all(v.is_assigned for v in self.__variables):
            return self(*self.__variables) 
        elif self.is_global:
            return self.__function(*self.__variables)
        else:
            return True
    
    def __repr__(self):
        return str(tuple([str(v.name) for v in self.variables]))

    @property
    def __xnary(self):
        return len(self.variables)
    
    @property
    def is_unary(self):
        return self.__xnary == 1
    
    @property
    def is_binary(self):
        return self.__xnary == 2
    
    @property
    def is_global(self):
        return self.__xnary > 2

    @property
    def arcs(self):
        if not self.is_binary:
            raise TypeError(f"{self}: constraint is not binary")
        L,R = self.variables
        return [(self, L, R), (self, R, L)]

    @property
    def variables(self):
        return self.__variables
    

class Constraint(Factor):
    pass


## NOT CURRENTLY VIABLE DO NOT USE
# from agent.environment import Environment 

# class __CSPEnvironment(Environment):
#     def __init__(self, problem, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.__csp = problem

#     @property
#     def csp(self):
#         return self.__csp

#     @property
#     def variables(self):
#         return self.csp.variables
    
#     @property
#     def constraints(self):
#         return self.csp.constraints

#     @property
#     def domains(self):
#         return self.csp.domains

#     @property
#     def is_done(self):
#         return all(variable.is_assigned for variable in self.variables) and \
#                  all(constraint.is_satisfied for constraint in self.constraints)




from ..agent.things import Thing, Agent, RationalAgent, ModelBasedAgent, Action
from numpy import inf

from typing import Collection, override


class MazeRunner(Agent):
    """ Agent that runs mazes, visualised as 👾 """
    @override
    def __repr__(self) -> str:
        return "👾"


class GoalBasedAgent(ModelBasedAgent):
    """ Base class for goal based agents 
    
        Requires implemnentation of methods:
            `program(percepts:Collection[Thing]) -> Action`
            `at_goal() -> bool`
    """
    @property
    def at_goal(self) -> bool:
        """ checks if agent is at goal, needs overriding """
        raise NotImplementedError


class UtilityBasedAgent(GoalBasedAgent):
    """ Base class for utility based agents 
    
        Requires implemnentation of methods:
            `program(percepts:Collection[Thing]) -> Action`
            `at_goal() -> bool`
            `utility(action:Action) -> float`
    """
    def maximise_utility(self, actions: Collection[Action]) -> Action:
        """ calculates maximum utility from list of actions 
        
        :param actions: collection of possible actions
        :return: action with maximum utility
        """
        max_u = -inf  # negative infinity
        _flag = False  # safety check
        for action in actions:
            u = self.utility(action)
            if u > max_u:
                max_u = u
                output = action
                _flag = True
        if not _flag: raise ValueError("No valid actions provided")
        
        return output  # type: ignore


    def utility(self, action: Action) -> float:
        """ calculates utility of an action, needs overriding """
        raise NotImplementedError



__all__ = ["colours", "fonts"]

from typing import Iterator, Iterable, TypeVar, override, Generic
from numbers import Complex

from collections import deque

Number = TypeVar("Number", int, float)
DataType = TypeVar("DataType")

class queue(Generic[DataType]):
    """ A simple queue class """
    def __init__(self) -> None:
        """ Initialise queue """
        self.data: deque[DataType] = deque()  # double-ended queue

    @override
    def __repr__(self) -> str:
        """ String representation of queue 

            Returns string of list of data in queue
        """
        return str(list(self.data))
 

    def __iter__(self) -> Iterator[DataType]:
        """ Iterator for queue """
        return iter(self.data)


    def push(self, x:DataType) -> None:
        """ Push item onto back of queue """
        self.data.append(x)


    def pop(self) -> DataType:
        """ Pop item from front of queue """
        return self.data.popleft()


class stack(queue[DataType]):
    """ A simple stack class """
    @override
    def pop(self) -> DataType:
        """ Pop item from top of stack """
        return self.data.pop()
    

def manhattan(a:tuple[Number, Number], b:tuple[Number, Number]) -> Number:
    """ Compute Manhattan distance between two (x, y) locations """
    ax, ay = a  # (x, y)
    bx, by = b  # (x, y)
    return abs(bx-ax) + abs(by-ay)


from collections.abc import Iterable
import warnings
from matplotlib import pyplot as plt
from collections import deque

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pygame

from ..engine import App
from ..util.colours import (
    COLOR_BLACK, COLOR_WHITE,
    COLOR_RED, COLOR_RED_DARK, COLOR_RED_LIGHT,
    COLOR_YELLOW, COLOR_BLUE, COLOR_GREEN_LIGHT
)

from .graph import Node, Graph, GraphEnvironment
from .things import Thing, Agent

PRESET_MAZES = {
    0: [" xxx o",
        "    x ",
        "x x   ",
        "  x xx"],
    1: [" xx  o",
        "    x ",
        "x x   ",
        "    xx"],
    2: ["           o",
        "x xxxxxxxxx ",
        "x x       x ",
        "x x xxxxx x ",
        "  x x     x ",
        "x   x xx xx ",
        "xxx x  xxxx ",
        "    xx      "],
    4: ["xxx                 xxxxxxxxx",
        "x   xxxxxxxxxxxxxxxxxxx   x x",
        "x xxxx                x x x x",
        "x xxxxxxxxxxxxxxxxxxx x x x x",
        "x                     x x x x",
        "xxxxxxxxxxxxxxxxxxxxx x x x x",
        "x   xx                x x x x",
        "x x xx xxx xx xxxxxxxxx x x x",
        "x x    x   xxox         x x x",
        "x x xx xxxxxxxxxxxxxxxx x x x",
        "xxx xx             xxxx x x x",
        "xxx xxxxxxxxxxxxxx xx x x x x",
        "xxx             xx    x x x x",
        "xxxxxx xxxxxxxx xxxxxxx x x x",
        "xxxxxx xxxx             x   x",
        "       xxxxxxxxxxxxxxxxxxxxxx"]
}

PRESET_STARTS = {
    0: (3, 0),
    1: (3, 0),
    2: (7, 0),
    4: (15, 0)
}

class MazeTile(Node):
    def __init__(self, passable:bool):
        super().__init__()
        self.is_passable = passable
        self.is_goal = False
        self.visited = False

    def __repr__(self):
        repr =  "⬜" if self.is_passable else "⬛"
        if hasattr(self, "location"):
            repr += str(self.location)
        return repr


class Maze(Graph):
    def __init__(self, template=None, size=None):
        super().__init__()
        if template is not None:
            if not self._is_valid_template(template):
                raise ValueError(f"{self}: template provided is not valid")
            self.template = template
            self.height, self.width = len(template), len(template[0])
            self.size = self.width, self.height
            self._generate_from_template()

    def __repr__(self):
        return "ꡙ‍"

    def _is_valid_template(self, template):
        def valid(row, size):
            return False if len(row) != size else not any(
                c not in (' ','o','x') for c in row)

        if isinstance(template, Iterable):
            if all(isinstance(row, Iterable) for row in template):
                width = len(template[0])
                return all(valid(row,width) for row in template)
        return False

    def _generate_from_template(self):
        self.grid = []
        w,h = self.width, self.height
        print(f"{self}: generating {w}x{h} maze from template")
        for i, row in enumerate(self.template):
            grid_row = []
            for j, tile in enumerate(row):
                node = MazeTile(tile != "x")  # True is passable
                if tile == "o":
                    node.is_goal = True
                node.location = (i, j)
                grid_row.append(node)
                if i > 0:
                    node.add_neighbour(self.grid[i-1][j])
                if j > 0:
                    node.add_neighbour(grid_row[j-1])
            self.grid.append(grid_row)
        self.add_node(self.grid[0][0])  # cascade adding nodes


    def plot_nodes(self, init=None):
        labels = {self.grid[i][j]: f"({i},{j})" 
                    for i in range(self.width)
                        for j in range(self.height)}
        condition = lambda node: node.is_passable
        return super().plot_nodes(init=init, labels=labels, condition=condition)


class MazeEnvironment(GraphEnvironment):
    def __init__(self, maze, *args, **kwargs):
        if not isinstance(maze, Maze):
            raise TypeError(f"{self}: maze must be valid Maze")
        super().__init__(maze, *args, **kwargs)
        self.size = self.width, self.height = maze.size
        self.success = False  # maze is solved
    def __repr__(self):
        return self.maze.__repr__()

    @property
    def maze(self):
        return self.graph

    @property
    def grid(self):
        return self.maze.grid

    @property
    def is_done(self):
        if len(self.agents) == 0: return True
        return self.success

    @property
    def goal(self):
        goal = {node for node in self.maze.nodes if node.is_goal}
        if len(goal) == 1:
            goal = next(iter(goal))
        return goal

    def show_graph(self, *args, **kwargs):
        return self.show(*args, **kwargs)

    def percept(self, agent):
        node = agent.location
        return node.neighbours

    def execute_action(self, agent, action):
        """ """
        command, node = action
        if command == "move":
            if node in self.maze:
                self.success = agent.move_to(node)
    
    def run(self, track_agent=False, *args, **kwargs):
        super().run(10000, *args, **kwargs)

    @classmethod
    def from_template(MazeEnvironment, template):
        maze = Maze(template)
        return MazeEnvironment(maze)


class GraphicMaze(MazeEnvironment):
    def run(self, graphical=True, lps=2, *args, **kwargs):
        if graphical:
            MazeApp(self, *args, name=f"{self}", lps=lps, **kwargs).run()
        else:
            super().run(*args, **kwargs)

  
class MazeApp(App):
    """ MazeApp
            Graphical version of Maze """
    def __init__(self, environment:MazeEnvironment=None, track_agent=False,**kwargs):
        if environment is None:
            print(f"{self}: building new MazeEnvironment from template")
            environment = MazeEnvironment.from_template(PRESET_MAZES[0])
        if not isinstance(environment, MazeEnvironment):
            raise TypeError(f"{self}: environment must be MazeEnvironment")
        
        self.environment = environment
        self._fit_to_environment()
        super().__init__(**kwargs)

        nx = self.environment.width
        fs = min(35, max(10, self.width//nx))
        self.thing_font = pygame.font.SysFont("segoe-ui-symbol", fs)
        self.counter = -1
        self._flag = True
        self.track = track_agent

    def _fit_to_environment(self):
        """ Fit width and height ratios to match environment """
        _flag = self.environment.width > self.environment.height
        if _flag:
            xy_ratio = self.environment.width/self.environment.height
            a, b = self.height, self.width
        else:
            xy_ratio = self.environment.height/self.environment.width
            b, a = self.height, self.width

        a = b // xy_ratio
        b = int(a*xy_ratio)
        # a = int(a)

        self.height, self.width = (a,b) if _flag else (b,a)
        self.size = self.width, self.height


    def update(self):
        if self.counter < 0:
            self.counter += 1
            return
        if not self.environment.is_done:
            self.environment.step()
            self.counter += 1
        elif self._flag:
            print(f"{self.environment}: Maze complete after {self.counter} iterations.")
            self._flag = False

    def render(self):
        # self.screen.fill(self.environment.color)
        self.render_grid()
        self.render_things()

    def render_grid(self):
        nx,ny = self.environment.width, self.environment.height
        self.tile_size = self.width / nx
        tile_origin = (0,0)
        grid = self.environment.grid
        tiles = []
        for i in range(ny):
            i_ = i# ny - i - 1
            
            row = []
            for j in range(nx):
                tileRect = pygame.Rect(
                    tile_origin[0] + j * self.tile_size,
                    tile_origin[1] + i * self.tile_size,
                    self.tile_size, self.tile_size)
                
                node = grid[i_][j]
                if node.is_goal:
                    color = COLOR_RED
                elif self.track and hasattr(node, "visited") and node.visited:
                    color = COLOR_GREEN_LIGHT
                elif node.is_passable:
                    color = COLOR_RED_LIGHT
                else:
                    color = COLOR_RED_DARK
                
                pygame.draw.rect(self.screen, color, tileRect)
                pygame.draw.rect(self.screen, COLOR_WHITE, tileRect, 1)
                row.append(tileRect)
            tiles.append(row)
        self.tiles = tiles
            
    def label_tile(self, node, string, color=COLOR_WHITE):
        i,j = node.location
        renderLoc = self.tiles[i][j].center
        thingRender = self.thing_font.render(string, True, color)
        thingRect = thingRender.get_rect()
        thingRect.center = renderLoc
        self.screen.blit(thingRender, thingRect)
    
    def render_things(self):
        locations = {}
        for node in self.environment.maze:
            if node.is_goal:
                self.label_tile(node, "👑", COLOR_YELLOW)

        for agent in self.environment.agents:
            location = agent.location
            self.label_tile(agent.location, str(agent), COLOR_BLUE)
            
            if self.track:
                if hasattr(agent, "frontier"):
                    for node in agent.frontier:
                        self.label_tile(node, "?", COLOR_WHITE)
                


from numpy import inf as infinity
from ..optimisation.things import *
from ..agent.environment import GraphicEnvironment
from ..search.util import manhattan
import random

PRESET_STATES = {
    "empty": None,
    '0': {
        "hospitals": [(2, 4)],
        "houses": [(2,1),(1,3),(8,0),(6,4)],
        "height": 5,
        "width": 10,
    },
    '1': {
        "hospitals": [(4,0), (9,3)],
        "houses": [(2,1),(1,3),(8,0),(6,4)],
        "height": 5, 
        "width": 10},
    '2': {
        "houses": [(2,1),(1,3),(8,0),(6,4)],
        "height": 5, 
        "width": 10},
    '3': {
        "houses": [(1, 2), (5, 1), (12, 2), (13, 2), (4, 0), (8, 5), (15, 1), (12, 4), (1, 7), (6, 5), (15, 3), (0, 1), (0, 0), (10, 5), (5, 3), (3, 6), (6, 3), (0, 7)],
        "hospitals": [(11,6), (5,2), (4,1)],
        "height": 8,
        "width": 16},
    '4': {
        "houses": [(1, 14), (17, 6), (2, 4), (4, 17), (9, 4), (3, 14), (8, 1), (17, 15), (6, 10), (9, 7), (3, 19), (10, 11), (8, 18), (0, 15), (4, 15), (4, 7), (16, 11), (5, 8), (13, 12), (8, 13), (17, 3), (1, 13), (11, 9), (10, 15), (2, 3), (7, 10), (5, 1), (14, 19), (14, 7), (18, 6), (8, 3), (11, 2), (5, 7), (11, 8), (18, 16), (12, 13), (5, 2), (18, 14)],
        "hospitals": [(8,0), (0, 12), (17, 12), (3,5)],
        "height": 20,
        "width": 20},
    '5': {
        "houses": [ (26, 22), (8, 13), (5, 29), (19, 15), (27, 13), (17, 24), (15, 6), (0, 11), (0, 26), (1, 22), (0, 14), (13, 17), (4, 10), (27, 10), (28, 28), (26, 29), (17, 28), (10, 18), (3, 8), (25, 6), (16, 28), (1, 15), (10, 23), (10, 6), (26, 15), (15, 12), (27, 12), (13, 1), (15, 10), (9, 29), (22, 9), (17, 29), (21, 24), (12, 17), (8, 15), (19, 13), (1, 26), (12, 18), (20, 3), (15, 18), (16, 27), (25, 18), (4, 17), (21, 13), (6, 29), (10, 25), (15, 23), (18, 12), (8, 20), (11, 18), (10, 19), (9, 26), (9, 23), (17, 1), (24, 10), (11, 4), (3, 3), (8, 12), (22, 10), (12, 21), (19, 28), (21, 29), (10, 17), (27, 1), (26, 9), (9, 14), (21, 15), (29, 18), (20, 15), (13, 12), (24, 17), (7, 20), (28, 4), (19, 17), (1, 8), (26, 26), (11, 28), (2, 1)],
        "hospitals": [(15, 5), (6, 1), (11, 3), (27, 22), (8, 14)],
        "height": 30,
        "width": 30},
}



class HospitalOptimiser(Optimiser, UtilityBasedAgent):
    def explore(self, state):
        if not state: return
        print(f"{self}: exploring state\n    {state['hospitals']}")
        for hospital, loc in state["hospitals"].items():
            hospital.location = loc

    def utility(self, state):
        """ calculate distance from each hospital to houses """
        obj = 0
        houses, hospitals = state["houses"], state["hospitals"]
        for house in houses:
            dist_to_nearest = infinity   # very big
            for hospital in hospitals:
                dist = manhattan(
                    houses[house],
                    hospitals[hospital])
                if dist < dist_to_nearest:
                    dist_to_nearest = dist
            obj += dist_to_nearest
        return -obj


class HospitalPlacement(GraphicEnvironment):
    def __init__(self, init=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.initialise_state(init)

    @property
    def state(self):
        return {
            "hospitals": {
                thing: thing.location
                    for thing in self.things
                        if isinstance(thing, Hospital)},
            "houses": {
                thing: thing.location
                    for thing in self.things
                        if isinstance(thing, House)},
            "bounds": {
                "xmin": 0, "xmax": self.width-1,
                "ymin": 0, "ymax": self.height-1}
        }

    @property
    def neighbours(self):
        neighbours = []
        for i, hospital in enumerate(self.state["hospitals"]):
            location = hospital.location
            for x,y in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                proposal = location[0] + x, location[1] + y
                if self.is_inbounds(proposal):
                    candidate = self.state.copy()
                    candidate["hospitals"][hospital] = proposal
                    neighbours.append(candidate)
        return neighbours

    def is_inbounds(self, location):
        if not super().is_inbounds(location):
            return False
        return len(self.things_at(location)) == 0

    def add_agent(self, agent:Agent):
        if not isinstance(agent, Agent):
            raise TypeError(f"{self}: {agent} is not an Agent.")
        self.agents.add(agent)

    def add_thing_randomly(self, thing):
        x = random.randint(self.x_start, self.x_end-1)
        y = random.randint(self.y_start, self.y_end-1)
        lim, count = 10, 0
        while not self.is_inbounds((x,y)):
            count += 1
            if count > lim:
                print(f"Tried and failed to add {thing} to environment")
                return     
            x = random.randint(self.x_start, self.x_end-1)
            y = random.randint(self.y_start, self.y_end-1)
        self.add_thing(thing, (x,y))


    def initialise_state(self, state_dict):
        if state_dict is None:
            return
        if "height" in state_dict:
            self.height = state_dict["height"]
        if "width" in state_dict:
            self.width = state_dict["width"]
        self.size = self.width, self.height
        if "hospitals" in state_dict:
            for loc in state_dict["hospitals"]:
                self.add_thing(Hospital(), location=loc)
        for loc in state_dict["houses"]:
            self.add_thing(House(), location=loc)
        

    @property
    def is_done(self):
        if len(self.agents) == 0: return True  # if there are no agents
        return hasattr(self, "success") and self.success

    def percept(self, agent):
        return self.state, self.neighbours

    def execute_action(self, agent, action):
        """ Execute an action """
        command, state = action
        match command:
            case "done":
                if state:
                    agent.explore(state)
                self.success = True
            case "explore":
                agent.explore(state)


import numpy as np

from .csp import *

class MagicTile(Variable):
    def __add__(self, y):
        x = self.value if self.value else 0
        y = (y.value if y.value else 0) if isinstance(y, Variable) else y
        return x+y
    
    def __radd__(self, y):
        return self + y  # symmetric add


class MagicSquare(ConstraintSatisfactionProblem):
    def __init__(self, size=4):
        variables, constraints = self.initialise_state(size)
        super().__init__(variables, constraints)

    @property
    def n(self):
        return self.__n
    
    @property
    def M(self):
        """ Magic constant """
        return self.n*(1+self.n**2)/2
    
    def __repr__(self):
        return str(self.__grid)

    def initialise_state(self, size):
        assert isinstance(size, int), f"{self}: value size ({size}) must be int"
        assert size > 0, f"{self}: size ({size}) must be greater than 0"
        self.__n = size
        domain = set(range(1,self.n**2+1))
        grid = np.matrix([
            [MagicTile(domain, str((i,j))) for j in range(self.n)] 
                for i in range(self.n)])
        self.__grid = grid
        variables = np.array(grid).ravel().tolist()
        
        def magic_constraint(*variables):
            return (np.sum(variables) <= self.M)

        row_constraints = {Factor(magic_constraint, grid[i,:])
                                for i in range(self.n)}
        col_constraints = {Factor(magic_constraint, grid[:,i])
                                for i in range(self.n)}
        diag_constraints = {Factor(magic_constraint, np.diag(grid)),
                            Factor(magic_constraint, np.diag(np.fliplr(grid)))}
        alldiff_constraint = {Factor(alldiff, variables)}

        constraints = row_constraints \
                        | col_constraints \
                            | diag_constraints \
                                | alldiff_constraint
        return  variables, constraints
        




# class __MagicSquareEnvironment(CSPEnvironment, GraphicEnvironment):
#     def __init__(self, n:int, *args, **kwargs):
#         csp = MagicSquare(n)
#         super().__init__(*args, problem=csp, height=n, width=n, **kwargs)
#         for i in range(n):
#             for j in range(n):
#                 self.add_thing(self.csp.state[i,j], location=(i,j))

#     @property
#     def is_done(self):
#         if len(self.agents) == 0: return True
#         return super().is_done

#     def percept(self, agent):
#         return self.csp
    
#     def add_agent(self, agent):
#         # self.things.add(agent)
#         self.agents.add(agent)

#     def execute_action(self, agent, action):
#         command, state = action

#         match command:
#             case "done":
#                 # should end naturally
#                 print("all good")
#             case "explore":
#                 agent.explore(self.csp)
#             case "backtrack":
#                 _variables = agent.backtrack()
#                 for variable, _variable in zip(self.csp.variables, _variables):
#                     variable.value = _variable.value



from collections.abc import Iterable
from collections import deque
from copy import deepcopy

import numpy as np

def alldiff(*variables):
    if len(variables) == 1:
        if not isinstance(variables[0], Iterable):
            return True
        variables = variables[0]
    # variables = [
    #     variable.value for variabl in variable if variable.is_assigned]
    values = [
        variable.value for variable in variables 
            if variable.is_assigned]
    return len(set(values)) == len(values)


def aslist(npcol):
    return np.array(npcol).ravel().tolist()


def revise(factor, A, B):
    is_revised = False
    if A.is_assigned: return False
    for value in A.domain.copy():
        A.value = value
        is_valid_B = False
        for _value in B.domain:
            B.value = _value
            if factor.is_satisfied:
                is_valid_B = True
            B.value = None
        if not is_valid_B:
            A.domain.remove(value)
            is_revised = True
        A.value = None
    return is_revised


def ac3(csp, log=False, inplace=True):
    if not inplace: csp = deepcopy(csp)
    arcs = deque(csp.arcs)
    while len(arcs) > 0:
        f, A, B = arcs.popleft()  # end of queue
        if log:
            print(f"considering arc from {A.name} to {B.name}")
            print(f"  before: {A.name} in {A.domain}, {B.name} in {B.domain}")
        if revise(f, A, B):
            if log:
                print(f"  after: {A.name} in {A.domain}, {B.name} in {B.domain}")
            if len(A.domain) == 0: return False if inplace else None
            for constraint in csp.constraints:
                if log:
                    print(f"  do we need to check constraint {constraint}")
                    print(f"   is binary? {constraint.is_binary}")
                    print(f"   contains {A.name}? {A in constraint}")
                    print(f"   doesn't contain {B.name}? {B not in constraint}")
                if constraint.is_binary \
                        and A in constraint\
                            and B not in constraint:
                    for arc in constraint.arcs:
                        if arc[-1] is A:
                            if arc not in arcs:
                                if log:
                                    print(f"    adding {arc} to arcs frontier")
                                arcs.append(arc)
                            elif log:
                                print(f"    arc {arc} already in frontier")
        elif log:
            print(f"  after: {A.name} in {A.domain}, {B.name} in {B.domain} (no change)")
    return True if inplace else csp


def make_node_consistent(csp, inplace=True):
    if not inplace: csp = deepcopy(csp)
    for variable in csp.variables:
        if variable.is_assigned: continue  # ignore any assigned variables
        domain = variable.domain.copy()  # copy this to avoid set size change errors 
        for value in domain:
            variable.value = value
            for constraint in csp.constraints:
                if constraint.is_unary and variable in constraint:
                    if not constraint.is_satisfied:
                        variable.domain.remove(value)
                        break
            variable.value = None
    if not inplace: return csp


from collections import deque

class queue:
    def __init__(self):
        self.data = deque()

    def __repr__(self):
        return str(list(self.data))
    
    def __iter__(self):
        return iter(self.data)

    def push(self, x):
        self.data.append(x)

    def pop(self):
        return self.data.popleft()

class stack(queue):
    def pop(self):
        return self.data.pop()
    

def manhattan(a, b):
    ax, ay = a  # (x, y)
    bx, by = b  # (x, y)
    return abs(bx-ax) + abs(by-ay)


"""THINGS.PY

Contains class definitions for some things
"""
from collections.abc import Callable, Collection, Iterable
from typing import override, Union

from ..util.fonts import platform

Action = Union[str, Iterable[str]]

class Thing:
    """The base class for all things"""
    @override
    def __repr__(self) -> str:
        return "❓" if platform != "darwin" else "?"

## Some physical things

class Obstacle(Thing):
    @override
    def __repr__(self) -> str:
        return "🚧" if platform != "darwin" else "X"


class Food(Thing):
    @override
    def __repr__(self) -> str:
        return "🍔" if platform != "darwin" else "f"


class Water(Thing):
    @override
    def __repr__(self) -> str:
        return "💧" if platform != "darwin" else "w"


class Animal(Thing):
    """ A generic animal thing """
    pass


class Dog(Animal):
    """If it looks like a dog and it barks like a dog ..."""
    @override
    def __repr__(self) -> str:
        return "🐶" if platform != "darwin" else "d"


## Some agent things

class Agent(Thing):
    """ Base class for all agents """
    pass


class RationalAgent(Agent):
    """ Base class for rational agent """
    def __init__(self, program:Callable) -> None:
        self.performance = 0
        # Check that program is callable function
        if program is None or not isinstance(program, Callable):
            raise ValueError("No valid program provided")
        self.program:Callable = program


class ModelBasedAgent(RationalAgent):
    """ Base class for model based agents 
    
    Requires a program method to be defined
        program(percepts:Collection[Thing]) -> Action
    """
    def __init__(self):
        """ Initialize the agent with its program """
        super().__init__(self.program)

    def program(self, percepts:Collection[Thing]) -> Action:
        """ The agent program 
        Given a collection of percepts, return an action
        """
        raise NotImplementedError

## State
class State(Thing):
    """ Base class for states """
    @override
    def __repr__(self) -> str:
        return self.__class__.__name__
    
class Bump(State):
    """ A bump state """
    pass


"""ENVIRONMENT.PY
    This contains code for agent environments and simulations
"""
import warnings
import math
import random
from collections.abc import Collection, Iterable
from typing import Union, TypeVar, override

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import pygame

from ..engine import App
from ..util.colours import COLOR_BLACK, COLOR_WHITE
from ..util.fonts import _get_symbolic_font_unsafe
from .things import Thing, Agent, Obstacle, Action

SYMBOL_FONT_NAME = _get_symbolic_font_unsafe()

T = TypeVar("T")
GenericLocation = T | tuple[T] | None
XYLocation = tuple[int, int]

class BaseEnvironment:
    """ Base Environment
            Adapted from AIMI code
    """
    def __repr__(self) -> str:
        """ Base string representation """
        return self.__class__.__name__
    
    @property
    def is_done(self) -> bool:
        """ Property indicating completion of simulation, needs overriding
        """
        raise NotImplementedError
    
    def step(self) -> None:
        """ Executes incremental step in discrete environment, needs overriding
        """
        raise NotImplementedError
    
    def run(self, steps:int = 100, pause_for_user:bool = True) -> None:
        """ Executes the simulation until either environment is done, or total
        steps exceeds limit
        
        :param          steps: max. # of iterations in simulation, default 100
        :param pause_for_user: require user input before beginning, default True
        """
        if pause_for_user:
            input("Press enter to start simulation")

        print(f"{self}: Running for {steps} iterations.")

        for i in range(steps):
            if self.is_done:  # print terination message and exit
                print(f"{self}: Simulation complete after {i} of {steps} iterations.")
                return
            self.step() # else iterate one step
        print(  # if loop completes, print max steps reached
            f"{self}: Simulation complete after {steps} of {steps} iterations.")


class Environment(BaseEnvironment):
    """ An Environment is a BaseEnvironment that has Things and Agents
    """
    def __init__(self):
        """ Initialises set of things and agents """
        self.things = set()  # all things in environment
        self.agents = set()  # all agents in environment

    @property
    def is_done(self) -> bool:
        """ Is considered done if there are no agents in environment """
        return len(self.agents) == 0  # cannot simulate with no agents


    def step(self) -> None:
        """ Executes percept, program and action step for each agent. """
        if not self.is_done:
            actions = {
                agent: agent.program(self.percept(agent))
                    for agent in self.agents}  # dictionary of actions by agents
            
            for agent, action in actions.items():
                # iterate through each action and allow agent to execute it
                self.execute_action(agent, action) 

            if self.is_done:
                print(f"{self}: Task environment complete. No further actions.")


    def percept(self, agent:Agent) -> Collection[Thing]:
        """ Returns the collection, e.g. set or list, of Things that an Agent 
        can perceive in the current state.
        
        :param        agent: the agent that is percieving
        """
        raise NotImplementedError


    def execute_action(self, agent:Agent, action:Action) -> None:
        """ For a given agent and action, performs execution
         
        :param        agent: the agent that is executing an action
        :param       action: the action statement to be executed by the agent
        """
        raise NotImplementedError


    def add_thing(self, thing:Thing, location:GenericLocation = None) -> None:
        """ Adds a thing to the environment at a given location (if applicable).
        
        :param        thing: the Thing (or Thing subclass) to be added
        :param     location: the location to add the Thing at (if applicable)
        """
        if isinstance(thing, type):  # if a class is given, instantiate it
            if issubclass(thing, Thing):
                thing = thing()
            else:  # not a Thing
                print(f"{self}: Tried to add {thing} but its not a Thing.")
                return

        if not isinstance(thing, Thing):  # not a Thing
            print(f"{self}: Tried to add {thing} but its not a Thing.")
            return
    
        if thing in self.things:  # prevent duplicate addition
            print(f"{self}: Tried and failed to add duplicate {thing}.")
            return
        
        if location:  # if a location is specified, set it
            thing.location = location  # set location of thing
            print(f"{self}: Adding {thing} at {location}")
            self.things.add(thing)  # add thing to environment

        if isinstance(thing, Agent):  # if thing is an agent, add to agents list
            print(f"{self}: Adding {thing} to list of agents.")
            self.agents.add(thing)
    

    def add_agent(self, agent:Agent, location:GenericLocation=None) -> None:
        """ Adds an agent to the environment at a given location 
        (if applicable).

        Wrapper for add_thing, specifically for agents.
        
        :param        agent: the Agent (or Agent subclass) to be added
        :param     location: the location to add the Agent at (if applicable)
        """
        if not isinstance(agent, Agent):  # not an Agent
            print(f"{self}: {agent} is not an Agent. Adding as Thing instead.")
        self.add_thing(agent, location)  # delegate to add_thing


    def delete_thing(self, thing:Thing) -> None:
        """ Removes a thing from the environment.
        
        :param        thing: the Thing (or Thing subclass) to be removed
        """
        if thing not in self.things: return  # thing not in environment

        self.things.remove(thing)  # remove thing from environment
        if isinstance(thing, Agent):  # if thing is an agent,
            self.agents.remove(thing)   # remove from agents list too


    def things_at(self, location:GenericLocation) -> list[Thing]:
        """ Returns list of things at a given location. 
        
        :param     location: the location to get Things at
        """
        return [thing for thing in self.things if thing.location == location]
    

    def __call__(self, location:GenericLocation) -> list[Thing]:
        """ Allows environment to be called as function to get things at location.
        
        Wrapper for things_at.

        :param     location: the location to get Things at
        """
        return self.things_at(location)


class XYEnvironment(Environment):
    """ A 2D grid environment for agents and things """
    DEFAULT_BG = COLOR_BLACK  # default background color

    def __init__(self, width:int = 10, height:int = 10) -> None:
        """ Initialises environment of given width and height
        
        :param        width: width of environment grid
        :param       height: height of environment grid
        """
        if not isinstance(width, int) or not isinstance(height, int):
            raise TypeError(f"{self}: dimensions must be integers")
        
        if width < 1 or height < 1:
            raise ValueError(f"{self}: dimensions must be greater than zero")
        
        super().__init__()  # initialise base environment

        if not hasattr(self, 'color'):  # set default color if not overridden
            self.color = self.DEFAULT_BG
        
        self.width, self.height = width, height  # set dimensions
        self.observers = []  # observers of environment changes
        self.bumped = set()  # agents that have bumped into obstacles
        self.x_start_offset = self.x_end_offset = 0  # offsets for walls
        self.y_start_offset = self.y_end_offset = 0  # offsets for walls
    

    @property
    def x_start(self) -> int:
        """ Starting x coordinate considering walls """
        return 0 + self.x_start_offset
    

    @property
    def x_end(self) -> int:
        """ Ending x coordinate considering walls """
        return self.width - self.x_end_offset
    

    @property
    def y_start(self) -> int:
        """ Starting y coordinate considering walls """
        return 0 + self.y_start_offset
    

    @property
    def y_end(self) -> int:
        """ Ending y coordinate considering walls """
        return self.height - self.y_end_offset


    def things_near(self, location:XYLocation, radius:int|float = 1) -> Collection[Thing]:
        """ Returns list of things near a given location within a certain radius.
        
        :param     location: the location to get Things near
        :param       radius: the radius around location to search
        """
        raise NotImplementedError
    

    def _add_wall(self, location:XYLocation) -> None:
        """ Adds a wall at the given location if no obstacle is present.
        
        :param     location: the location to add wall at
        """
        class Wall(Obstacle): pass  # simple wall class

        if all(not isinstance(obj, Obstacle)  # only add wall if no obstacle
               for obj in self.things_at(location)):
            self.add_thing(Wall(), location)  # add wall to environment


    def add_walls(self) -> None:
        """ Adds walls around the perimeter of the environment """
        if self.width > 2:  # add left and right walls
            for y in range(self.height):
                self._add_wall((0, y))
                if self.width > 1: self._add_wall((self.width-1, y))

        if self.height > 2:  # add top and bottom walls
            for x in range(self.width):
                self._add_wall((x, 0))
                if self.height > 1: self._add_wall((x, self.height - 1))
            self.y_start_offset += 1  # account for walls
            self.y_end_offset += 1  # account for walls

        if self.width > 2:
            self.x_start_offset += 1  # account for walls
            self.x_end_offset += 1  # account for walls
    

    def is_valid(self, location:XYLocation) -> bool:
        """ Checks if a location is valid in the environment
        
        :param     location: the location to validate
        """
        if not isinstance(location, Iterable): return False  # must be iterable
        if len(location) != 2:  return False  # must be length 2
        if any(map(lambda x: not isinstance(x,int), location)): return False  # must be ints
        return True
    
    def is_inbounds(self, location):
        if not self.is_valid(location): return False
        x,y = location
        if not (x >= self.x_start and x < self.x_end): return False
        if not (y >= self.y_start and y < self.y_end): return False
        return True

    @override
    def add_thing(self, thing:Thing, location:XYLocation) -> None:
        if location is None:  # default to starting location
            location = (self.x_start, self.y_start)
        elif self.is_inbounds(location):  # check location validity
            location = tuple(location)  # force tuple to make hashable
        else:  # invalid location
            print(f"Tried and failed to add {thing} to environment")
            return
        
        if isinstance(self, Agent):  # if thing is an agent
            thing.bump = False  # give capacity to be bumped
        super().add_thing(thing, location)  # delegate to base method

        assert thing in self.things_at(location)  # sanity check


    def add_thing_randomly(self, thing:Thing):
        """ Adds a thing to a random location in the environment
        
        :param        thing: the Thing (or Thing subclass) to be added
        """
        x = random.randint(self.x_start, self.x_end-1)  # random x coord
        y = random.randint(self.y_start, self.y_end-1)  # random y coord
        self.add_thing(thing, (x,y))  # delegate to add_thing


class GraphicEnvironment(XYEnvironment):
    """ A graphical 2D grid environment for agents and things """
    @override
    def run(self, graphical=True, steps=100, **kwargs) -> None:
        """ Executes the simulation in graphical or base mode.
        
        :param       graphical: whether to run graphical version
        :param           steps: max. # of iterations in simulation, default 100
        :param        **kwargs: additional arguments for environment base class
        """
        if graphical:  # run graphical version
            EnvironmentApp(self, steps=steps, name=f"{self}", **kwargs).run()
        else:  # run base version
            super().run(steps=steps, **kwargs)


class EnvironmentApp(App):
    """EnvironmentApp
            Graphical version of an XYEnvironment 
    """
    # size = width, height = 600, 400  # uncomment to override
    def __init__(self,
                 environment:XYEnvironment|None = None, 
                 steps:int = 100, 
                 **kwargs) -> None:
        """ Initialises graphical environment app
        
        :param   environment: the XYEnvironment to visualise
        :param         steps: max. # of iterations in simulation, default 100
        :param      **kwargs: additional arguments for App base class
        """
        if environment is None:  # default to 12x8 XYEnvironment
            print(f"{self}: No environment specified, using default")
            environment = XYEnvironment(12, 8)
        if not isinstance(environment, XYEnvironment):  # check type
            raise TypeError(f"{self}: environment must be XYEnvironment")

        self.environment = environment
        self._fit_to_environment()  # fit window aspect to environment
        super().__init__(**kwargs)  # initialise base App class

        # font for rendering things
        self.thing_font = pygame.font.SysFont(SYMBOL_FONT_NAME, 28)

        self.counter = -1  # start at -1 to allow initial render
        self.steps = steps  # max. steps in simulation
        self._flag = True  # flag for completion message
        
    
    def _fit_to_environment(self) -> None:
        """ Fit width and height ratios to match environment """
        # determine whether environment is wider than tall
        _wider_than_tall = self.environment.width > self.environment.height

        if _wider_than_tall: 
            xy_ratio = self.environment.width/self.environment.height
            a, b = self.height, self.width
        else:
            xy_ratio = self.environment.height/self.environment.width
            b, a = self.height, self.width

        a = b // xy_ratio  # adjust dimensions
        b = int(a*xy_ratio)

        self.height, self.width = (a,b) if _wider_than_tall else (b,a)
        self.size = self.width, self.height  # set size attribute


    @override
    def update(self) -> None:
        """ Main process loop 
                Steps the environment simulation
        """
        if self.counter < 0:  # initial render only
            self.counter += 1
            return
        if self.counter < self.steps and not self.environment.is_done:
            self.environment.step()
            self.counter += 1
        elif self._flag:  # print completion message once
            print(f"{self.environment}: Simulation complete after {self.counter} of {self.steps} iterations.")
            self._flag = False


    @override
    def render(self) -> None:
        """ Main render loop 
                Renders the environment grid and things
        """
        self.screen.fill(self.environment.color)  # write background color
        self.render_grid()  # render grid
        self.render_things()  # render things


    def render_grid(self) -> None:
        """ Render environment grid """
        nx,ny = self.environment.width, self.environment.height
        self.tile_size = self.width // nx  # size of each tile
        tile_origin = (0, 0)

        tiles:list[list[pygame.Rect]] = []  # 2D list of tile rects

        for i in range(ny):  # for each row
            row = [] 
            for j in range(nx):  # for each column
                tileRect = pygame.Rect(  # define tile rectangle
                    tile_origin[0] + j * self.tile_size,
                    tile_origin[1] + i * self.tile_size,
                    self.tile_size, self.tile_size)
                # draw tile border
                pygame.draw.rect(self.screen, COLOR_WHITE, tileRect, 1)
                # add tile render object to row
                row.append(tileRect)
            tiles.append(row)  # add row to tiles

        self.tiles = tiles  # store tiles for later use


    def render_things(self) -> None:
        """ Render things in environment """
        locations:dict = {}  # map of locations to things

        # build location map
        for thing in self.environment.things:
            if thing.location in locations:
                locations[thing.location].append(thing)
            else:
                locations[thing.location] = [thing]

        # render things at each location
        for location, things in locations.items():
            n_things = len(things)
            if n_things > 1:  # multiple things, pick one at random per render
                thing = things[random.randint(0,n_things-1)]
            elif n_things == 1:  # single thing
                thing = things[0]
            else:  # no things to render at this location
                continue
            # determine render location
            renderLoc = self.tiles[location[1]][location[0]].center
            thingRender = self.thing_font.render(str(thing), True, COLOR_WHITE)
            thingRect = thingRender.get_rect()
            thingRect.center = renderLoc
            # blit thing to screen
            self.screen.blit(thingRender, thingRect)


from .logic import *

from copy import copy

def is_literal(x: Proposition) -> bool:
    return type(x) in [Proposition, Not]

class KnowledgeBase(Proposition):
    """ Represents the knowledge base of an agent """
    def __init__(self, *args):
        match len(args):
            case 0:
                self._conjunction =  Sentence()
            case 1:
                self._conjunction = args[0]
            case _:
                self._conjunction: Conjunction = Conjunction(*args)
        self._entails = set()

    def __repr__(self) -> str:
        return f"KB: {self._conjunction}"


    def evaluate(self, model: dict) -> bool:
        """ evalutes a model over the knowledge base """
        return self._conjunction.evaluate(model)
    
    def __call__(self, model: dict) -> bool:
        """ evaluates a model """
        return self.evaluate(model)

    @property
    def symbols(self) -> set:
        """ returns the set of symbols that exist in the knowledge base """
        return self._conjunction.symbols

    def to_cnf(self) -> None:
        """ converts knowledge base into conjunctive normal form """
        self._conjunction = cnf(self._conjunction)
        
    def as_cnf(self):
        """ returns a new knowledge base in conjunctive form """
        KB = copy(self)
        KB.to_cnf()
        return KB

    def __invert__(self) -> Proposition:
        if isinstance(self._conjunction, Proposition):
            return KnowledgeBase(~self._conjunction)
        else:
            return self
    
    def __and__(self, x:Proposition) -> Proposition:
        if not isinstance(x, Proposition):
            raise TypeError(f"input must be Proposition, was type {type(x)}")
        if isinstance(self._conjunction, Proposition):
            return Conjunction(self._conjunction, x)
        else:
            return x

    def __or__(self, x:Proposition) -> Proposition:
        if not isinstance(x, Proposition):
            raise TypeError(f"input must be Proposition, was type {type(x)}")
        if isinstance(self._conjunction, Proposition):
            return Disjunction(self._conjunction, x)
        else:
            return x

    def __contains__(self, x:Proposition) -> bool:
        """ returns if symbol is in knowledge base 
             - symbols and their negations (P, ~P only)
        """
        return (x in self._conjunction)
    
    def entails(self, x:Proposition) -> bool:
        """ returns if knowledge base entails proposition """
        if is_literal(x) and x in self._entails: return True
        entailment = model_check(self, x)
        if entailment and (type(x) in [Proposition, Not]):
            self._entails.add(x)
        return entailment
    
    def models(self, x:Proposition) -> bool:
        """ alias for entails """
        return self.entails(x)

    def _add(self, x: Proposition, validate: bool = False) -> None:
        """ add a new proposition to the knowledge base """
        if not isinstance(x, Proposition):
            raise TypeError(f"input must be Proposition, was type {type(x)}")
        if isinstance(self._conjunction, Proposition):
            if validate and self.entails(~x):
                raise ValueError(
                    f"cannot add contradiction to Knowledge Base, KB ⊨ {~x}")
                return
            self._conjunction &= x
        else:
            self._conjunction = x  # add symbol for first time

    def add(self, *xs:tuple[Proposition], validate: bool = False) -> None:
        """ add new proposition(s) to the knowledge base """
        for x in xs:
            self._add(x, validate)  # add other arguments



def model_check(knowledge:KnowledgeBase, sentence:Proposition):
    """ run model checking algorithm """
    symbols = set.union(knowledge.symbols, sentence.symbols)
    return _check_all(knowledge, sentence, symbols, {})



def _check_all(knowledge, query, symbols, model):
    """
        enumerates all models in knowledge base and 
        checks if query is entailed
    """
    if not symbols:
        if knowledge.evaluate(model):
            return query.evaluate(model)
        return True
    else:
        remaining = symbols.copy()
        p = remaining.pop()

        model_true = model.copy()
        model_true[p] = True
        model_false = model.copy()
        model_false[p] = False
    return (
        _check_all(knowledge, query, remaining, model_true)
             and _check_all(knowledge, query, remaining, model_false))



from ..agent.things import Thing


#= Formatting
def _balanced(s:str) -> bool:
    """ checks whether a string has balanced parantheses """
    counter:int = 0
    for c in s:
        if c == "(":
            counter += 1
        elif c == ")":
            if counter <= 0:
                return False
            counter -= 1
    return counter == 0


def format(formula: str) -> str:
    """ formats a formula by wrapping in parantheses if not already.
        will not put parenthesis around symbols or their negation
    """
    if not len(formula) or formula.isalpha() or (
        formula[0] == "(" and formula[-1] == ")" and _balanced(formula[1:-1])
    ) or (formula[0] == "¬" and formula[1:].isalpha()):
        return formula
    else:
        return f"({formula})"
    

#= Classes
class Sentence(Thing):
    """ base class of a sentence, empty """
    def __repr__(self) -> str:
        return self.formula
    

    def evaluate(self, model: dict) -> bool:
        return True

    @property
    def formula(self) -> str:
        return ""


    def __contains__(self, x) -> bool:
        return False


    @property
    def symbols(self) -> set:
        return set()

    
class Proposition(Sentence):
    """ Logical proposition representing logical statement """
    def __init__(self, name:str) -> None:
        self._name:str = name

    
    def __eq__(self, x:Sentence) -> bool:
        return (self is x)
    
    
    def __hash__(self) -> int:
        return hash(self.formula)
    
    
    def __bool__(self) -> None:
        """ prevents evaluation of boolean logic using, e.g. and and or, to avoid confusion """
        raise TypeError(f"cannot evaluate {type(self)}: use Proposition.evaluate(model)")
    
    @property
    def formula(self) -> str:
        """ the string representation of the proposition """
        return self._name
    
    @property
    def symbols(self) -> set:
        """ set of logical symbols in proposition """
        return {self}
    
    
    def __call__(self, model) -> bool:
        return self.evaluate(model)
    
    def evaluate(self, model: dict) -> bool:
        """ evaluates a model containing self """
        if self in model:
            return bool(model[self])
        else:
            raise IndexError(f"{self} not in model {model}")

               
    def __contains__(self, x:Sentence) -> bool:
        if type(x) not in (Proposition,Not):
            raise TypeError(
                f"Cannot evaluate if type {type(x).__name__} is in {type(self).__name__}")
        return (x in self.symbols) if type(x) is Proposition else False
        
    
    def __invert__(self) -> Sentence:
        """ returns negation of self (¬self)"""
        return Not(self)
    
    
    def __and__(self, x:Sentence) -> Sentence:
        """ returns conjunction (self ∧ input) """
        return And(self, x)
    
    
    def __or__(self, x:Sentence) -> Sentence:
        """ returns disjunction (self ∨ input) """
        return Or(self, x)
    
    def onlyif(self, x:Sentence) -> Sentence:
        """ 'self only if input', converse of 'input if self'
            returns implication (self -> input)
        """
        return Implication(self, x)
    
    def implies(self, x:Sentence) -> Sentence:
        """ returns implication (self -> input)"""
        return Implication(self, x)

    def iff(self, x:Sentence) -> Sentence:
        """ 'if and only if' , returns biconditional (self <-> input)"""
        return Biconditional(self, x)

                             
class Not(Proposition):
    """ Logical proposition representing negation or 'not' """
    def __init__(self, symbol: Proposition):
        super().__init__("¬"+format(symbol.formula))
        self.operand:Proposition = symbol
    
    
    def evaluate(self, model: dict) -> bool:
        """ evaluates a model containing operand """
        return not self.operand(model)

    @property
    def symbols(self) -> set:
        return self.operand.symbols
    
    
    def __contains__(self, x: Proposition) -> bool:
        if type(x) is Not and type(x.operand) is Proposition:
            return x.operand == self.operand
        return x in self.operand

    
    def __invert__(self) -> Sentence:
        """ returns ¬(¬operand), so operand"""
        return self.operand


class Conjunction(Proposition):
    """ Logical conjunction (∧) of two or more propositions """
    def __init__(self, operand0:Proposition, operand1:Proposition, *operands):
        _operands:list = self._expand(operand0) + self._expand(operand1)
        for operand in operands:
            _operands.extend(self._expand(operand))

        formula:str = format(_operands[0].formula)
        for operand in _operands[1:]:
            formula += " ∧ " + format(operand.formula)

        super().__init__(formula)
        self.operands:tuple = tuple(_operands)
        
    def _expand(self, operand:Proposition) -> list:
        """ returns individual operands in a conjunction
            (P ∧ Q) becomes [P, Q]
            R becomes [R]
        """
        if isinstance(operand, Conjunction):
            return list(operand.operands)
        else:
            return [operand]
    
    
    def __and__(self, operand: Proposition):
        return Conjunction(*self.operands, operand)

    
    def evaluate(self, model: dict) -> bool:
        return all(operand.evaluate(model) for operand in self.operands)
    
    
    def __contains__(self, x:Proposition) -> bool:
        return any(x in operand for operand in self.operands)

    @property
    def symbols(self) -> set:
        return set.union(*[operand.symbols for operand in self.operands])

And = Conjunction  # alias


class Disjunction(Proposition):
    """ Logical disjunction (∨) of two or more propositions """
    def __init__(self, operand0:Proposition, operand1:Proposition, *operands):
        
        _operands:list = self._expand(operand0) + self._expand(operand1)
        for operand in operands:
            _operands.extend(self._expand(operand))

        formula:str = format(_operands[0].formula)
        for operand in _operands[1:]:
            formula += " ∨ " + format(operand.formula)

        super().__init__(formula)
        self.operands:tuple = tuple(_operands)

        
    def _expand(self, operand:Proposition) -> list:
        """ returns individual operands in a disjunction
            (P ∨ Q) becomes [P, Q]
            R becomes [R]
        """
        if isinstance(operand, Disjunction):
            return list(operand.operands)
        else:
            return [operand]
    
    
    def __or__(self, operand: Proposition):
        return Disjunction(*self.operands, operand)
    
    
    def evaluate(self, model: dict) -> bool:
        return any(operand.evaluate(model) for operand in self.operands)
    
    
    def __contains__(self, x:Proposition) -> bool:
        return any(x in operand for operand in self.operands)

    @property
    def symbols(self) -> set:
        return set.union(*[operand.symbols for operand in self.operands])
    
Or = Disjunction  # alias


class Implication(Proposition):
    """ Logical implication (->) of two propositions """
    def __init__(self, left:Proposition, right:Proposition):
        self.antecedent:Proposition = left
        self.consequent:Proposition = right
        self.left:Proposition = self.antecedent
        self.right:Proposition = self.consequent

    @property
    def formula(self): 
        return f"{format(self.antecedent.formula)} -> {format(self.consequent.formula)}"
    
    
    def evaluate(self, model: dict) -> bool:
        return ((not self.antecedent.evaluate(model)) or 
                    self.consequent.evaluate(model))

    @property
    def symbols(self) -> set:
        return set.union(self.antecedent.symbols, self.consequent.symbols)
    

class Biconditional(Implication):
    """ Logical biconditional (<->) of two propositions """
    @property
    def formula(self):
        return f"{format(self.left.formula)} <-> {format(self.right.formula)}"

    
    def evaluate(self, model: dict):
        left = self.left.evaluate(model)
        right = self.right.evaluate(model)
        return (left and right) or (not left and not right)
    

#= Utility functions
def simplify(statement:Proposition) -> Proposition:
    """ Simplifies propositions into logical ands and ors"""
    match statement:
        case Biconditional():
            l, r = statement.antecedent, statement.consequent
            return simplify(l.implies(r) & r.implies(l))
        case Implication():
            return simplify(~statement.antecedent | statement.consequent)
        case Not():
            proposition = simplify(statement.operand)
            match proposition:
                case Conjunction() | Disjunction():
                    return Conjunction(*[~p for p in proposition.operands])
                case _:
                    return ~proposition
        case Conjunction():
            out = Conjunction(
                *[simplify(p) for p in statement.operands if p is not None])
            return out
        case Disjunction():
            out = Disjunction(
                *[simplify(p) for p in statement.operands if p is not None])
            return out
        case _:
            return statement


def distribute(sentence:Disjunction) -> Sentence:
    """ Distributes disjunctions """
    if type(sentence) is Conjunction:
        return Conjunction(
            *[distribute(operand) for operand in sentence.operands])
    if type(sentence) is not Disjunction:
        return sentence
    if not any([type(p) is Conjunction for p in sentence.operands]):
        return sentence
    distr = []
    for proposition in sentence.operands:
        if type(proposition) is Conjunction:
            for operand in proposition.operands:
                conj = Disjunction(operand,*[p for p in sentence.operands if p is not proposition])
                distr.append(conj)
    return Conjunction(*distr)


def cnf(sentence:Sentence) -> Sentence:
    """ returns the conjunctive normal form of a logical sentence"""
    return distribute(simplify(sentence))



COLOR_BLACK = (0,0,0)
COLOR_WHITE = (255,255,255)

# RGB
## RED
COLOR_RED = (255, 0, 0)
COLOR_RED_DARK  = (100, 0, 0)
COLOR_RED_LIGHT = (255, 100, 100)
## GREEN
COLOR_GREEN = (0, 255, 0)
COLOR_GREEN_DARK  = (0, 100, 0)
COLOR_GREEN_LIGHT = (100, 255, 100)
COLOR_GREEN_GRASS = (80, 240, 100) 
## BLUE
COLOR_BLUE = (0, 0, 255)
COLOR_BLUE_DARK  = (0, 0, 100)
COLOR_BLUE_LIGHT = (100, 100, 255)

# CYM
## CYAN
COLOR_CYAN = (0, 255, 255)
COLOR_CYAN_DARK  = (0, 100, 100)
COLOR_CYAN_LIGHT = (100, 255, 255)
## YELLOW
COLOR_YELLOW = (255, 255, 0)
COLOR_YELLOW_DARK  = (100, 100, 0)
COLOR_YELLOW_LIGHT = (255, 255, 100)
## MAGENTA
COLOR_MAGENTA = (255, 0, 255)
COLOR_MAGENTA_DARK  = (100, 0, 100)
COLOR_MAGENTA_LIGHT = (255, 100, 255)


from setuptools import setup

setup(
    name="co2114",
    version="2026.0.3",
    description="codebase for co2114",
    author="wil ward",
    python_requires=">=3.12",
    packages=[
        "co2114",
        "co2114.agent", 
        "co2114.search", 
        "co2114.optimisation",
        "co2114.constraints", 
        "co2114.constraints.csp",
        "co2114.reasoning",
        "co2114.util"],
    install_requires=[
        'pygame>=2.6', 
        'numpy', 
        'ipython',
        'matplotlib',
        'jupyter',
        'scikit-learn',
        'seaborn'
        ]
)



from ..agent.things import Agent, RationalAgent
from ..agent.environment import Environment

from .logic import *
from .inference import *

import random
from collections import deque
from typing import Union, TypeVar
T = TypeVar("T")


def permute_indices(n: int, i: int) -> tuple[int]:
    """ returns ordered permutation of indices from 0 to n-1 starting at i """
    if i < 0 or i >= n:
        raise ValueError(f"i must be in [0, {n-1}] for n = {n}. i = {i}")
    return (i, *list(range(i+1,n)), *list(range(0, i)))

def choice(bag: set[T]) -> T:
    """ returns a random element from a set """
    # return next(iter(bag))
    return random.choice(list(bag))

def sample(bag: set[T], n: int) -> set[T]:
    """ returns a sample of _n_ items from a set """
    if n < 1 or n > len(bag):
        raise ValueError(f"n must be between 1 and {len(bag)}, n = {n}")
    i, _sample = 0, set()
    while i < n:
        _sample.add(choice(bag - _sample))
        i+=1
    return _sample


WEAPON_NAMES = {
    "candlestick", "dagger", 
    "lead piping", "revolver", 
    "rope", "spanner"
}

CHARACTER_NAMES = {
    "scarlett", "mustard",
    "white", "green",
    "peacock", "plum"
}

ROOM_NAMES = {
    "kitchen", "ballroom", "conservatory",
    "dining room", "billiard room", "library",
    "study", "hall", "lounge"
}


def create_symbols() \
        -> tuple[set[Proposition],set[Proposition],set[Proposition]]:
    """ returns set of proposition for cluedo game (characters, rooms, weapons)
    """
    # characters_names = 

    characters : set[Proposition] = {
        Proposition(name) for name in CHARACTER_NAMES}
    rooms      : set[Proposition] = {
        Proposition(name) for name in ROOM_NAMES     }
    weapons    : set[Proposition] = {
        Proposition(name) for name in WEAPON_NAMES   }
    return characters, rooms, weapons


class LogicalAgent(RationalAgent):
    knowledge : KnowledgeBase = KnowledgeBase()


Card = Proposition
CluedoSymbols = dict[str, set[Card]] 
Guess = tuple[
        Union[Card,None],
        Union[Card,None],
        Union[Card,None]
    ]  # type alias


class CluedoAgent(RationalAgent):
    """ a cluedo player, but not necessary a logical one """
    def __init__(self, symbols: CluedoSymbols, name: Union[str, None] = None):
        self.symbols : CluedoSymbols = symbols
        self.__name = name

    def guess(self) -> Guess:
        """ makes a random guess (character, room, weapon) """
        _guess = []
        for key in ["characters", "rooms", "weapons"]:
            card = choice(self.symbols[key])
            _guess.append(card)

        return tuple(_guess)

    def receive_hand(self, hand: set[Card]) -> None:
        """ receives hand of cards """
        self.hand : set[Card] = hand

    def reveal_in_hand(self, guess: Guess) -> Union[Card, None]:
        """ reveals card in hand if guessed """
        if not hasattr(self, "hand"):
            raise KeyError(f"player {self} has no hand")
        for symbol in guess:
            if symbol in self.hand:
                return symbol  # return symbol

    def program(self,
                percepts : tuple[Agent, Guess, Union[Card, None]]
            ) -> Union[Guess, Card, None]:
        """ reveals a card if in hand or makes a guess """
        agent, guess, revealed = percepts

        if agent is not self:  # not my turn
            if revealed is None:  # noone revealed anything yet
                return self.reveal_in_hand(guess)
            return None  # does not contain in hand
        elif all(card is not None for card in guess) :
            return  # already made a guess this turn
        return self.guess()  # random guess
    
    def __repr__(self) -> str:
        return self.__name if self.__name else super().__repr__()


# class CluedoLogicalAgent(CluedoAgent, LogicalAgent):
#     """ a logical cluedo player """
#     def __init__(self, symbols: CluedoSymbols):
#         super().__init__(symbols)

#         # Must be at least one character, one room and one weapon
#         self.knowledge.add(
#             Disjunction(*self.symbols["characters"]),
#             Disjunction(*self.symbols["rooms"]),
#             Disjunction(*self.symbols["weapons"])) 

#         self.solution: Guess = (None, None, None)
#         self.is_not: set[Card] = set()


#     def receive_hand(self, hand: set[Card]) -> None:
#         """ receives hand of cards """
#         super().receive_hand(hand)
#         for card in self.hand:
#             self.knowledge.add(~card)  # solution cannot cards in hand
#             self.is_not.add(card)  # set of ruled out cards
#         print(f"{self.knowledge}")


#     @property
#     def insight(self) -> str:
#         outstr = f"{self}:\n"
#         for key in ["characters", "rooms", "weapons"]:
#             outstr += f"{key}:" + "{\n"
#             for symbol in self.symbols[key]:
#                 if self.knowledge.entails(symbol):
#                     outstr += f"\t{symbol}: YES!\n"
#                 elif self.knowledge.entails(~symbol):
#                     outstr += f"\t{symbol}: No.\n"
#                 else:
#                     outstr += f"\t{symbol}: maybe...\n"
#             outstr += "      }\n"
#         return outstr

#     def guess(self) -> Guess:
#         valid_characters = [
#             card for card in self.symbols["characters"] 
#                 if card not in self.is_not]
#         valid_rooms = [
#             card for card in self.symbols["rooms"] 
#                 if card not in self.is_not]
#         valid_weapons = [
#             card for card in self.symbols["weapons"] 
#                 if card not in self.is_not]
#         guess = (random.choice(valid_characters),
#                  random.choice(valid_rooms),
#                  random.choice(valid_weapons))
#         return guess

#     def program(self,
#                 percepts: tuple[Agent, Guess, Union[Card,None]]
#             ) -> Union[Guess, Proposition, None]:
#         """ agent program takes agent, proposition and returns a guess
#             (character, room, weapon)
#         """
#         agent, guess, revealed = percepts
#         if all(card is not None for card in guess):
#             if revealed is not None:
#                 if agent is self:  # if revealed to agent
#                     print(f"{self}: adding {~revealed} to knowledge")
#                     self.knowledge.add(~revealed)
#                     self.is_not.add(revealed)
#                 else:              # unseen reveal
#                     character, room, weapon = guess
#                     clause = ~character | ~room | ~weapon  # one must be wrong !
#                     print(f"{self}: adding {clause} to knowledge")
#                     self.knowledge.add(clause)
#             elif agent is self:  # no card was revealed on own guess
#                 character, room, weapon = guess
#                 if character not in self.hand:
#                     print(f"{self} : adding {character} to knowledge")
#                     self.solution=(character,self.solution[1],self.solution[2])
#                     self.knowledge.add(character)
#                     self.is_not |= self.symbols["characters"] - {character}
#                 if room not in self.hand:
#                     print(f"{self} : adding {room} to knowledge")
#                     self.solution = (self.solution[0], room, self.solution[2])
#                     self.knowledge.add(room)
#                     self.is_not |= self.symbols["rooms"] - {room}
#                 if weapon not in self.hand:
#                     print(f"{self} : adding {weapon} to knowledge")
#                     self.solution = (self.solution[0], self.solution[1], weapon)
#                     self.knowledge.add(weapon)
#                     self.is_not |= self.symbols["weapons"] - {weapon}

#             if agent is not self:
#                 if revealed is None:
#                     return self.reveal_in_hand(guess)
#                 return None  # does not contain in hand

#         print(f"{self}: checking knowledge")
#         if self.solution[0] is None:
#             for symbol in self.symbols["characters"] - self.is_not:
#                 if self.knowledge.entails(symbol):  # character must be in solution
#                     self.solution = (symbol, self.solution[1], self.solution[2])
#                     self.is_not |= self.symbols["characters"] - {symbol}
#                     break
#         if self.solution[1] is None:
#             for symbol in self.symbols["rooms"] - self.is_not:
#                 if self.knowledge.entails(symbol):  # room must be in solution
#                     self.solution = (self.solution[0], symbol, self.solution[2])
#                     self.is_not |= self.symbols["rooms"] - {symbol}
#                     break
#         if self.solution[2] is None:
#             for symbol in self.symbols["weapons"] - self.is_not:
#                 if self.knowledge.entails(symbol):  # weapon must be in solution
#                     self.solution = (self.solution[0], self.solution[1], symbol)
#                     self.is_not |= self.symbols["weapons"] - {symbol}
#                     break
    
#         if any(card is None for card in self.solution):
#             return self.guess()  # makes a guess
#         else:
#             return self.solution

#     def __repr__(self) -> str:
#         return "🕵️"


class CluedoGame(Environment):
    def __init__(self, partial: bool = False):
        super().__init__()

        characters, rooms, weapons = create_symbols()
        if partial:
            characters = sample(characters, 4)
            rooms      = sample(rooms,      4)
            weapons    = sample(weapons,    4)
        
        self.cards   : dict[str, Card]          = {
                "characters": characters,
                "rooms": rooms,
                "weapons": weapons
            }
        self.players  : list[CluedoAgent] = []
        self.player   : int               = -1
        self.guess    : Guess             = (None, None, None)
        self.revealed : Union[Card, None] = None

        
    def __repr__(self) -> str:
        return "♟️"
        
    def add_agent(self, agent: CluedoAgent, *args, **kwargs) -> None:
        """ adds agent to player list and environment """
        if agent in self.players:
            raise ValueError(f"agent {agent} already signed up to play")
        self.players.append(agent)
        return super().add_agent(agent, *args, **kwargs)


    def add_player(self, player: CluedoAgent) -> None:
        """ adds player, alias for add_agent """
        return self.add_agent(player)

    @property
    def deck(self) -> set[Card]:
        """ returns all cards in the deck """
        return set.union(*self.cards.values())


    def deal(self) -> None:
        """ deal cards to agents """
        n_players = len(self.players)
        deck = deque(self.deck - set(self.__solution))
        hands = {player: set() for player in range(n_players)}
        player = 0
        random.shuffle(deck)
        while len(deck) > 0:
            hands[player].add(deck.pop())
            player = (player + 1) % n_players
        print(f"dealing hands {hands}")
        for player in range(n_players):
            self.players[player].receive_hand(hands[player])


    def step(self) -> None:
        """ a single turn in cluedo """
        n_players = len(self.players)
        
        if self.player < 0:  # first turn
           self.__solution = (
                choice(self.cards["characters"]),
                choice(self.cards["rooms"]),
                choice(self.cards["weapons"])
            )
           self.deal()
        self.player = (self.player + 1) % n_players
        self.guess, self.revealed = (None, None, None), None
        
        order = permute_indices(n_players, self.player)
        for i in order:
            agent = self.players[i]
            self._take_turns(agent)
            if self.revealed is not None:
                for _agent in self.players:
                    if isinstance(_agent, LogicalAgent) and  \
                            (agent is not _agent 
                                or agent is not self.players[self.player]):
                        self._take_turns(_agent)
                break  # card revealed, show to player
        # reveal card, if any, to player
        agent = self.players[self.player]
        self._take_turns(agent)

        if self.is_done:
            print(f"{self}: {agent} won with guess {self.guess}")
            print(f"{self}: Task environment complete. No further actions.")
            return

    def _take_turns(self, agent: CluedoAgent):
        print(f"{agent} <- {self.state}")
        action = agent.program(self.percept(agent))
        self.execute_action(agent, action)

    def percept(self, agent: CluedoAgent):
        return self.state
    
    def execute_action(self, 
                       agent: CluedoAgent,
                       action: Union[Guess, Card, None]) -> None:
        match action:
            case tuple():  # Guess
                print(f"{agent} guessed {action}")
                self.guess = action
            case Proposition():  # Card
                print(f"{agent} revealed {action} to {self.state[0]}")
                self.revealed = action
            case None: 
                if agent is not self.state[0]:  # no card revealed
                    print(f"{agent} didn't have a card to show {self.state[0]}")
            case _:
                raise TypeError(f"action {action} had invalid type, must be Union[tuple, Proposition, None]")

    @property
    def state(self) -> tuple[CluedoAgent, Guess, Union[Card,None]]:
        return (self.players[self.player], self.guess, self.revealed)

    @property
    def is_done(self) -> bool:
        try:
            return (self.guess == self.__solution) | super().is_done
        except:
            return False


if __name__ == "__main__":
    environment = CluedoGame(partial=True)
    cards = environment.cards
    player0 = CluedoAgent(cards, "0")
    # player1 = CluedoLogicalAgent(cards)
    player2 = CluedoAgent(cards, "2")
    player3 = CluedoAgent(cards, "3")

    environment.add_player(player0)
    # environment.add_player(player1)
    environment.add_player(player2)
    # environment.add_player(player3)

    environment.run()